@@ -155,6 +155,7 @@ type error =
   | Binary
   | Connection_failure of string
   | Unexpected_status of result_status * string * (result_status list)
+  | Cancel_failure of string
 
 (** [Error error] indicates an [error] *)
 exception Error of error
@@ -168,14 +169,6 @@ exception Oid of oid
 
 (** {6 Utility functions}*)
 
-val escape_string : ?pos : int -> ?len : int -> string -> string
-(** [escape_string pos len str] escapes ASCII-substring [str] of length
-    [len] starting at position [pos] for use within SQL.
-
-    @param pos default = 0
-    @param len default = String.length str - pos
-*)
-
 external unescape_bytea : string -> string = "PQunescapeBytea_stub"
 (** [unescape_bytea str] unescapes binary string [str]. *)
 
@@ -717,7 +710,7 @@ object
   (** [request_cancel] requests that PostgreSQL abandon processing of
       the current command.
 
-      @raise Error if there is a connection error.
+      @raise Error if there is a connection or cancel error.
   *)
 
 
@@ -801,6 +794,17 @@ object
       @raise Error if there is a connection error.
   *)
 
+
+  (** Escaping *)
+
+  method escape_string : ?pos : int -> ?len : int -> string -> string
+  (** [escape_string ?pos ?len str] escapes ASCII-substring [str]
+      of length [len] starting at position [pos] for use within SQL.
+
+      @param pos default = 0
+      @param len default = String.length str - pos
+  *)
+
   method escape_bytea : ?pos : int -> ?len : int -> string -> string
   (** [escape_bytea ?pos ?len str] escapes binary substring [str]
       of length [len] starting at position [pos] for use within SQL.
